# MIT License

# Copyright (c) 2024 Benedikt Burger

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .main_window import Functions
from .mix_methods import MixMethods


class TwoDMixSP(MixMethods):
    """The '2D-mix-SP' method."""

    _function = Functions.TWOD_MIX_SP
    _accept_pos = (506, 600)
    _run_pos = (140, 200)
    _close_pos = (540, 140)
    _change_inputs_pos = (150, 266)
    _result_pos = (133, 293)

    # coordinates of the 2DmixLP-function (in FHD standard)
    _configuration_pos = {
        "Wavelengths (nm)": [(400, 186), (460, 186), (520, 186)],
        "Indexes of refraction": [(400, 200), (460, 200), (520, 200)],
        "Group velocity index": [(400, 213), (460, 213), (520, 213)],
        "Group delay dispersion": [(400, 226), (460, 226), (520, 226)],
        "Phase (radians)": [(400, 246), (460, 246), (520, 246)],
        "Input face reflectivity": [(400, 260), (460, 260), (520, 260)],
        "Output face reflectivity": [(416, 280), (483, 280), (550, 280)],
        "Crystal absorption (per mm)": [(400, 293), (460, 293), (520, 293)],
        "n2 red1 (sq cm/W)": [(400, 306), (460, 306), (520, 306)],
        "n2 red2 (sq cm/W)": [(400, 326), (460, 326), (520, 326)],
        "n2 blue (sq cm/W)": [(400, 340), (460, 340), (520, 340)],
        "beta red1 (cm/W)": [(400, 360), (460, 360), (520, 360)],
        "beta red2 (cm/W)": [(400, 373), (460, 373), (520, 373)],
        "beta blue (cm/W)": [(400, 393), (460, 393), (520, 393)],
        "Pulse energy (Joules)": [(400, 406), (460, 406), (520, 406)],
        "Pulse duration (ps)": [(400, 420), (460, 420), (520, 420)],
        "Pulse Delay (ps)": [(400, 440), (460, 440), (520, 440)],
        "Pulse chirp (THz/ps)": [(400, 453), (460, 453), (520, 453)],
        "Beam diameter (mm)": [(400, 473), (460, 473), (520, 473)],
        "Supergaussian coefficient": [(400, 486), (460, 486), (520, 486)],
        "Walkoff angle (mrad)": [(400, 500), (460, 500), (520, 500)],
        "Beam position (mm)": [(400, 520), (460, 520), (520, 520)],
        "Radius of curvature (mm)": [(400, 533), (460, 533), (460, 533)],
        "Number t,x,y points": [(400, 546), (460, 546), (520, 546)],
        "Size of crystal/grid (mm)": [(400, 566), (460, 566), (520, 566)],
        "deff (pm/V)": [(400, 580)],
        "delta k (1/mm)": [(400, 596)],
        "Number of z steps": [(400, 612)],
        "Dist. to detector (mm)": [(400, 628)]
    }



# MIT License

# Copyright (c) 2024 Benedikt Burger

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .main_window import Functions
from .mix_methods import MixMethods


class TwoDCavLP(MixMethods):
    """The '2D-cav-LP' method."""

    _function = Functions.TWOD_CAV_LP
    _accept_pos = (480, 537)
    _run_pos = (148, 275)
    _result_pos = (148, 357)
    _change_inputs_pos = (157, 310)
    _close_pos = (480, 135)
    _configuration_pos = {"Wavelengths (nm)": [(330, 151), (397, 151), (464, 151)],
                          "Indexes of refraction": [(330, 167), (397, 167), (464, 167)],
                          "Crystal left reflectivity": [(330, 183), (397, 183), (464, 183)],
                          "Crystal right reflectivity": [(330, 199), (397, 199), (464, 199)],
                          "Crystal loss (per mm)": [(330, 215), (397, 215), (464, 215)],
                          "Enrgy/Pwr left (J/W)": [(330, 231), (397, 231), (464, 231)],
                          "Enrgy/Pwr right (J/W)": [(330, 247), (397, 247)],
                          "Pulse duration (FWHM ns)": [(330, 263), (397, 263), (464, 263)],
                          "Pulse delay (ns)": [(330, 279), (397, 279)],
                          "Beam diameter (FWHM mm)": [(330, 295), (397, 295), (464, 295)],
                          "Supergaussian coefficient": [(330, 311), (397, 311), (464, 311)],
                          "Walk off angle (mard)": [(330, 327), (397, 327), (464, 327)],
                          "Beam offset (mm)": [(397, 343), (464, 343)],
                          "Beam radius of curv. (mm)": [(330, 359), (397, 359), (464, 359)],
                          "Left mirror reflectivity": [(330, 375), (397, 375), (464, 375)],
                          "Right mirror reflectivity": [(330, 391), (397, 391), (464, 391)],
                          "Phase L-C (radians)": [(330, 407), (397, 407), (464, 407)],
                          "Phase C-R (radians)": [(330, 423), (397, 423), (464, 423)],
                          "Phase R-L (radians)": [(330, 439), (397, 439), (464, 439)],
                          "Mirror roc L R (mm)": [(330, 463), (397, 463)],
                          "Dist. L-C C-R R-L (mm)": [(330, 479), (397, 479), (464, 479)],
                          "Grid numbers z x y": [(330, 495), (397, 495), (464, 495)],
                          "Crystal/grid sizes (mm)": [(330, 511), (397, 511), (464, 511)],
                          "Cavity type/inversion": [(330, 527), (397, 527)],
                          "Deff (pm/V)/delta k (1/mm)": [(330, 543), (397, 543)]}

    def interpret_results(self, rows: list[str]) -> dict[str, float | list[float]]:
        """Interpret the results."""
        return {
            "Right input (W W -)": [float(i) for i in rows[0].split()[6:]],
            "Left input (W W J)": [float(i) for i in rows[1].split()[6:]],
            "Left output energy (J)": [float(i) for i in rows[2].split()[5:]],
            "Right output energy (J)": [float(i) for i in rows[3].split()[5:]],
            "So (W/sq cm)": float(rows[4].split()[4]),
        }



# MIT License

# Copyright (c) 2024 Benedikt Burger

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any
from collections import defaultdict

from .base_function import BaseFunction, Functions


class PW_OPO_BB(BaseFunction):
    _function = Functions.PW_OPO_BB
    _run_pos = (433, 475)
    _result_pos = (182, 618)
    _close_pos = (515, 174)
    _configuration_pos = {
        "Wavelengths (nm)": [(351, 219), (423, 219), (489, 219)],
        "Index of refraction": [(351, 235), (423, 235), (489, 235)],
        "Group velocity index": [(351, 251), (423, 251), (489, 251)],
        "Crystal left reflectivity": [(351, 267), (423, 267), (489, 267)],
        "Crystal right reflectivity": [(351, 283), (423, 283), (489, 283)],
        "Crystal loss (per mm)": [(351, 299), (423, 299), (489, 299)],
        "Enrgy/Pwr left (J/W)": [(351, 315), (423, 315), (489, 315)],
        "Enrgy/Pwr right (J/W)": [(351, 331), (423, 331)],
        "Pulse duration (FWHM ns)": [(351, 347), (423, 347), (489, 347)],
        "Beam diam. (FWHM mm)": [(351, 363), (423, 363), (489, 363)],
        "Left mirror reflectivity": [(351, 379), (423, 379), (489, 379)],
        "Right mirror reflectivity": [(351, 395), (423, 395), (489, 395)],
        "L mir. to Xtal phase (rad)": [(351, 411), (423, 411), (489, 411)],
        "R mir. to Xtal phase (rad)": [(351, 427), (423, 427), (489, 427)],
        "Cavity/Xtal length (mm)": [(351, 443), (423, 443)],
        "Pump bandwidth (MHz)": [(351, 459)],
        "Pump mode spacing (MHz)": [(351, 475)],
        "Pump FM (1/0)": [(351, 491)],
        "Deff (pm/V)": [(351, 507)],
        "Cav. type (1=lin. 0=ring)": [(351, 523)],
        "Delta k(1/mm)": [(351, 539)],
        "# of z integration steps": [(351, 555)],
        "New noise (1/0)": [(351, 571)]
    }

    def interpret_results(self, rows: list[str]) -> dict[str, Any]:
        """Interpret the results and return them as a dictionary."""
        data = defaultdict(list) # to support duplicate rows e.g., in qmix for different phase-matching conditions
        for row in rows:
            # skip empty lines
            if not row:
                continue
            if row == 'ERROR: No phase match found.':
                raise Exception('No phase match found')
            text, values = row.split("=")
            content = []
            for element in values.split():
                try:
                    val = float(element)
                except ValueError:
                    val = element
                content.append(val)
            data[text.strip()].append(content)
        return data

    def run_and_read(
        self,
        waiting_time: float = 0.1,
        max_tries: int = 10,
        interval: float = 0.1,
        waiting_line_count: int = 3,
    ) -> dict[str, Any]:
        return super().run_and_read(waiting_time, max_tries, interval, waiting_line_count)



# MIT License

# Copyright (c) 2024 Benedikt Burger

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The SNLO main window
"""

from typing import Optional, TypeVar

from .utils import gui, scale, get_screenfactors, set_screenfactors, Point
from .functions import Functions, open_function
from .base_function import BaseFunction
from .ref_index import RefractiveIndex
from .focus import Focus
from .two_d_mix_lp import TwoDMixLP
from .two_d_mix_sp import TwoDMixSP
from .q_mix import QMix
from .pw_opo_bb import PW_OPO_BB
from .two_d_cav_lp import TwoDCavLP


function_classes = {
    Functions.REF_INDEX: RefractiveIndex,
    Functions.QMIX: QMix,
    Functions.TWOD_MIX_LP: TwoDMixLP,
    Functions.TWOD_MIX_SP: TwoDMixSP,
    Functions.PW_OPO_BB: PW_OPO_BB,
    Functions.TWOD_CAV_LP: TwoDCavLP,
    Functions.FOCUS: Focus,
}

FunctionClass = TypeVar("FunctionClass")


class MainWindow:
    _close_pos = (95, 14)

    def __init__(self, screenfactors: Optional[Point] = None, **kwargs) -> None:
        super().__init__(**kwargs)
        sf = get_screenfactors()
        if sf is None or screenfactors is not None:
            set_screenfactors(new_factors=screenfactors)

    def close(self) -> None:
        gui.click(*scale(self._close_pos))

    def open_function(self, key: str | Functions) -> Optional[BaseFunction]:
        open_function(key)
        if key in function_classes:
            return function_classes[key]()

    def _open_function(self, function_class: type[FunctionClass]) -> FunctionClass:
        function = function_class()
        function.open()  # type: ignore
        return function

    def open_refractive_index(self) -> RefractiveIndex:
        return self._open_function(RefractiveIndex)

    def open_q_mix(self) -> QMix:
        return self._open_function(QMix)

    def open_two_d_mix_lp(self) -> TwoDMixLP:
        return self._open_function(TwoDMixLP)

    def open_two_d_mix_sp(self) -> TwoDMixSP:
        return self._open_function(TwoDMixSP)

    def open_pw_opo_bb(self) -> PW_OPO_BB:
        return self._open_function(PW_OPO_BB)

    def open_two_d_cav_lp(self) -> TwoDCavLP:
        return self._open_function(TwoDCavLP)

    def open_focus(self) -> Focus:
        return self._open_function(Focus)



# MIT License

# Copyright (c) 2024 Benedikt Burger

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re

import numpy as np


def import_snlo_file(file_name: str, file_path: str = "C:/SNLO/"):
    """import a file generated by SNLO, specified by filename and return array (no header)"""
    with open(file_path + file_name, "r") as f:
        input = f.readlines()
    data = []
    e_numbers = re.compile(r"-*[\d.]*?E-*[+]*\d*")
    r""" Regex
        -*        "-" or nothing
        [\d.]*?   decimal number (0 or more characters)
        -*        "-" or nothing
        [+]*      "+" or nothing
        \d*       decimal number (0 or more characters)
    """
    for i in input:
        temp = []
        # old implementation, suffers from strings without whitespace, e.g.:
        #    '9.206897E-100-1.616264E-2'
        #    element = i.split()
        #    for item in element:
        #        temp.append(float(item))
        new_element = e_numbers.findall(i)
        for item in new_element:
            temp.append(float(item))
        data.append(temp)
    return np.array(data.copy())



# MIT License

# Copyright (c) 2024 Benedikt Burger

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Autoclicker setup
=================

General methods for the autoclicker
"""

import logging
from typing import Any, Optional, Union

import pyautogui as gui
from pyperclip import paste


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

Point = Union[gui.Point, tuple[float, float]]


"""

Setup of the screen and scaling
-------------------------------

All positions in code are given on a Full HD (1920 * 1080) screen and dynamically adjusted to the
screen resolution.
"""


def read_display_screenfactors(standard: Point = (1920, 1080)) -> Point:
    """Read the scaling factor from Full HD to the current display resolution."""
    width, height = gui.size()
    return standard[0] / width, standard[1] / height


def set_screenfactors(new_factors: Optional[tuple[float, float]] = None) -> tuple[float, float]:
    """Set the screenfactors to `new_factors` or detect them automatically."""
    global factors
    factors = read_display_screenfactors() if new_factors is None else new_factors
    return factors


def get_screenfactors() -> Optional[tuple[float, float]]:
    """Get the current screenfactors or None, if not yet set."""
    global factors
    try:
        return factors
    except NameError:
        return None


def scale(x: Union[float, Point], y: Optional[float] = None) -> Point:
    """Scale coordinates from the definition standard to the current screen."""
    global factors
    if isinstance(x, (list, tuple)):
        if y is None:
            x, y = x
        else:
            raise ValueError("You cannot specify x as a tuple and y.")
    elif y is None:
        raise ValueError("You have to specify two coordinatres.")
    try:
        return x / factors[0], y / factors[1]
    except NameError:
        log.warning("Factors was not set, running `set_screenfactors` to get values.")
        set_screenfactors()
        return x / factors[0], y / factors[1]


def standard_position() -> Point:
    """Get the mouse position in standard coordinates (x, y)."""
    point = gui.position()
    global factors
    return point.x * factors[0], point.y * factors[1]


"""
Helper functions
----------------

GUI functions to get/set content from/into data fields.
"""


def get_content(position: Point) -> str:
    """Get the content of the field at position via double click.

    If there is a "-" in the text, the extraction fails!
    """
    gui.doubleClick(*scale(*position))
    gui.hotkey("ctrl", "c")
    return paste()


def get_value(position: Point) -> float:
    """Move to position, retrieve value and return float."""
    return float(get_content(position))


def get_content_complete(position: Point) -> str:
    """Go to position and retrieve the content there, marking all."""
    gui.click(*scale(*position))
    gui.hotkey("ctrl", "home")
    # both shift keys are necessary if keylock is on
    gui.hotkey("ctrl", "shiftleft", "shiftright", "end")
    gui.hotkey("ctrl", "c")
    return paste()


def get_value_complete(position: Point) -> float:
    """Move to position, retrieve value via context menu (slower) and return float."""
    return float(get_content_complete(position))


def set_value(position: Point, value: Any) -> None:
    """Move to position, insert value as string."""
    gui.doubleClick(*scale(*position))
    gui.press("delete")
    gui.doubleClick()
    gui.write(str(value))


def alt_tab() -> None:
    gui.hotkey("alt", "tab")


def get_position() -> Point:
    return gui.position()



# MIT License

# Copyright (c) 2024 Benedikt Burger

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any
from collections import defaultdict

from .base_function import BaseFunction, Functions


class QMix(BaseFunction):
    _function = Functions.QMIX
    _run_pos = (174, 61)
    _result_pos = (160, 260)
    _close_pos = (446, 20)
    _configuration_pos = {
        "Crystal": [(250, 76)],
        "Temperature": [(360, 76)],
        "Red1": [(155, 155)],
        "Red2": [(209, 155)],
        "Blue": [(266, 155)],
        "XY": [(357, 128)],
        "XZ": [(396, 128)],
        "YZ": [(429, 128)],
        "Mix": [(359, 168)],
        "OPO": [(416, 168)]
    }

    def interpret_results(self, rows: list[str]) -> dict[str, Any]:
        """Interpret the results and return them as a dictionary."""
        data = defaultdict(list) # to support duplicate rows e.g., in qmix for different phase-matching conditions
        for row in rows:
            # skip empty lines
            if not row:
                continue
            if row == 'ERROR: No phase match found.':
                raise Exception('No phase match found')
            text, values = row.split("=")
            content = []
            for element in values.split():
                try:
                    val = float(element)
                except ValueError:
                    val = element
                content.append(val)
            data[text.strip()].append(content)
        return data

    def run_and_read(
        self,
        waiting_time: float = 0.1,
        max_tries: int = 10,
        interval: float = 0.1,
        waiting_line_count: int = 3,
    ) -> dict[str, Any]:
        return super().run_and_read(waiting_time, max_tries, interval, waiting_line_count)



# MIT License

# Copyright (c) 2024 Benedikt Burger

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any

from .base_function import BaseFunction, Functions


class RefractiveIndex(BaseFunction):
    _function = Functions.REF_INDEX
    _run_pos = (295, 187)
    _result_pos = (160, 260)
    _close_pos = (365, 41)
    _configuration_pos = {
        "Crystal": [(170, 90)],
        "Temperature": [(290, 90)],
        "theta": [(170, 146)],
        "phi": [(300, 146)],
        "Wavelength": [(170, 190)],
    }

    def run_and_read(
        self,
        waiting_time: float = 0.1,
        max_tries: int = 10,
        interval: float = 0.1,
        waiting_line_count: int = 3,
    ) -> dict[str, Any]:
        return super().run_and_read(waiting_time, max_tries, interval, waiting_line_count)

    def refractive_indices(
        self, Crystal=None, Temperature=None, theta=None, phi=None, Wavelength=None
    ) -> list[float]:
        """Get the refractive indices (o, e).

        For the crystal, you have to use letters to press, for example "BB" to select the second
        crystal starting with a "B". In order to ensure, that it starts correctly, use any other
        letter first, e.g. "ABB".
        """
        kwargs = {
            "Crystal": Crystal,
            "Temperature": Temperature,
            "theta": theta,
            "phi": phi,
            "Wavelength": Wavelength,
        }
        results = self.configure_run_read(kwargs)
        return results["Refractive index (o,e)"]



# MIT License

# Copyright (c) 2024 Benedikt Burger

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional

from .base_function import BaseFunction
from .functions import open_function, Functions
from .utils import Point, gui, scale, set_value, get_value_complete


# coordinates of the Focus-function (in FHD standard)
_dict_focus: dict[str, Point] = {
    "Wavelength (nm)": (366, 220),
    "Refractive Index": (366, 240),
    "Waist size (mm)": (366, 260),
    "Face to focus (mm)": (366, 290),
    "Dist. to focus (mm)": (366, 300),
    "Rayleigh z in xtal (mm)": (366, 353),
    "Beam size (mm)": (366, 380),
    "Radius of curv. (mm)": (366, 393),
    "Far field ang air (mrad)": (366, 413),
    "fwhm": (219, 216),
    "1e^2": (166, 216),
}


def focus(
    wavelength_nm: float, ref_index: float, fwhm_mm: float, focus_pos_mm: float
) -> tuple[float, float, float, float]:
    """Call 'focus' function to determine parameters for the radius of curvature.

    inputs: wavelength_nm, ref_index, waist_size_mm (FWHM), focus_pos_mm
    return: zr, diameter (FWHM), radcurv, angle
    """
    open_function(Functions.FOCUS)
    gui.leftClick(*scale(*_dict_focus["fwhm"]))
    set_value(_dict_focus["Wavelength (nm)"], wavelength_nm)
    set_value(_dict_focus["Refractive Index"], ref_index)
    set_value(_dict_focus["Waist size (mm)"], fwhm_mm)
    set_value(_dict_focus["Face to focus (mm)"], focus_pos_mm)
    # Setting 'Dist. to focus (mm)' equal to 'Face to focus (mm)' gives parameters in air at the
    # input face
    set_value(_dict_focus["Dist. to focus (mm)"], focus_pos_mm)
    # readout
    zr = get_value_complete(_dict_focus["Rayleigh z in xtal (mm)"])
    diameter = get_value_complete(_dict_focus["Beam size (mm)"])
    radcurv = get_value_complete(_dict_focus["Radius of curv. (mm)"])
    angle = get_value_complete(_dict_focus["Far field ang air (mrad)"])
    return zr, diameter, radcurv, angle


class Focus(BaseFunction):
    _function = Functions.FOCUS

    def __init__(self) -> None:
        super().__init__()
        self._configuration_pos = {key: [value] for key, value in _dict_focus.items()}

    def read_results(self) -> list[str]:
        return super().read_results()

    def focus(
        self,
        wavelength_nm: Optional[float] = None,
        ref_index: Optional[float] = None,
        fwhm_mm: Optional[float] = None,
        focus_pos_mm: Optional[float] = None,
    ) -> tuple[float, float, float, float]:
        self.configure(
            {
                "Wavelength (nm)": wavelength_nm,
                "Refractive Index": ref_index,
                "Waist size (mm)": fwhm_mm,
                # Set face to focus and dist to focus to same value gives values in air at face
                "Face to focus (mm)": focus_pos_mm,
                "Dist. to focus (mm)": focus_pos_mm,
            }
        )
        zr = get_value_complete(_dict_focus["Rayleigh z in xtal (mm)"])
        diameter = get_value_complete(_dict_focus["Beam size (mm)"])
        radcurv = get_value_complete(_dict_focus["Radius of curv. (mm)"])
        angle = get_value_complete(_dict_focus["Far field ang air (mrad)"])
        return zr, diameter, radcurv, angle



# MIT License

# Copyright (c) 2024 Benedikt Burger

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from enum import StrEnum

from .utils import Point, gui, scale


# coordinates of the functions (in FHD standard)
_functions_coord: dict[str, Point] = {
    "Ref. Ind.": (66, 46),
    "Qmix": (66, 66),
    "Bmix": (66, 93),
    "QPM": (66, 120),
    "Opoangles": (66, 146),
    "Ncpm": (66, 173),
    "GVM": (66, 200),
    "PW-mix-LP": (66, 233),
    "PW-mix-SP": (66, 260),
    "PW-mix-BB": (66, 286),
    "2D-mix-LP": (66, 313),
    "2D-mix-SP": (66, 340),
    "PW-cav-LP": (66, 366),
    "PW-OPO-SP": (66, 393),
    "PW-OPO-BB": (66, 420),
    "2D-cav-LP": (66, 446),
    "Focus": (66, 473),
    "Cavity": (66, 500),
}


class Functions(StrEnum):
    """Enum for the functions."""

    REF_INDEX = "Ref. Ind."
    QMIX = "Qmix"
    BMIX = "Bmix"
    QPM = "QPM"
    OPO_ANGLES = "Opoangles"
    NCPM = "Ncpm"
    GVM = "GVM"
    PW_MIX_LP = "PW-mix-LP"
    PW_MIX_SP = "PW-mix-SP"
    PW_MIX_BB = "PW-mix-BB"
    TWOD_MIX_LP = "2D-mix-LP"
    TWOD_MIX_SP = "2D-mix-SP"
    PW_CAV_LP = "PW-cav-LP"
    PW_OPO_SP = "PW-OPO-SP"
    PW_OPO_BB = "PW-OPO-BB"
    TWOD_CAV_LP = "2D-cav-LP"
    FOCUS = "Focus"
    CAVITY = "Cavity"


def open_function(key: str | Functions) -> None:
    """opens function according to key"""
    gui.click(*scale(*_functions_coord[key]))



# MIT License

# Copyright (c) 2024 Benedikt Burger

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Example of a simulation using SNLO

A hash followed by two or more percent signs (e.g. "# %%") indicates cells to execute in ipython.

1. Start SNLO (do not move it!)
2. Execute the cells.
   If a cell starts with `alt_tab`, SNLO should be the program run last
"""

# %%
# Imports

import matplotlib.pyplot as plt
import numpy as np
from scipy import constants as cs

from snlohelper.main_window import MainWindow
from snlohelper import utils

# open the class for the main window
mw = MainWindow()


# %%
"Hard facts"
"""
Values which are given by physics or by the whole setup
"""

# Wavelengths etc.
signal_wl = 3534 / 3  # nm == 1178 nm
pump_wl = 532  # nm
temperature = 300  # K

# Result of qmix for the crystal (BBO)
"""
1178.0(o)+  532.0(o)=  366.5(e)
Walkoff [mrad]   =     0.00   0.00  69.90
Phase velocities = c/  1.652  1.674  1.667
Group velocities = c/  1.672  1.722  1.769
GrpDelDisp(fs^2/mm) =   32.7  136.0  229.4
At theta             =   29.8    deg.
Deff                 =  2.03E0   pm/V
S_o × L^2            =  2.35E7   Watt
Crystal ang. tol.×L  =    0.31   mrad°cm
Temperature range×L  =   17.39   K°cm
Mix accpt ang×L =     1.02    0.45 mrad°cm
Mix accpt bw×L  =   310.43  640.98 GHz°cm
"""
upconverted_wl = 366.5  # resulting wavelength
theta = 29.8
Deff = 2.03e0
walkoff = (0, 0, 69.9)  # angles in mrad
length = 12  # mm


# %%%
# calculate the refractive indices

def get_n():
    utils.alt_tab()
    ref = mw.open_refractive_index()
    s, _ = ref.refractive_indices("Ab", Temperature=temperature, theta=theta, Wavelength=signal_wl)
    p, _ = ref.refractive_indices(Wavelength=pump_wl)
    _, u = ref.refractive_indices(Wavelength=upconverted_wl)
    ref.close()
    return (s, p, u)


n = get_n()

# %%%
# Laser system

# fwhm pulse duration in ns
signal_ns = 3.3
pump_ns = 6

# fwhm beam diameter at entry face in mm (for elliptical 1st in walkoff, second orthogonal)
signal_mm = 1.3647e-1
signal_curv_mm = 3.4981e2
pump_mm = "0.914 0.397"
pump_curv_mm = "-1.7e3 -3.34E3"
upconverted_mm = 0.07788
upconverted_curv_mm = 7.4392e2

# grid size: length, width_walkoff, height_orthogonal, all in mm
grid = (length, 2.5, 1)


# %%
"Simulation settings"

signal_photons = 100_000
pump_E = 1e-9  # in J

# Calculate
signal_E = signal_photons * cs.h * cs.c / signal_wl / 1e-9  # in J
walkoff_offset = (0, 0, walkoff[2] * -1e-3 * length / 2)  # to have overlap in the center


# %%%
# Configure SNLO initially

utils.alt_tab()
mix = mw.open_two_d_mix_lp()

# start configuration: set all values in order to have a predefined configuration
mix.configure({
    "Wavelengths (nm)": (signal_wl, pump_wl, upconverted_wl),
    "Indexes of refraction": n,
    "Phases at input (rad)": [0.0, 0.0, 0.0],
    "Input face reflectivity (0-1)": [0.0, 0.0, 0.0],
    "Output face reflectivity (0-1)": [0.0, 0.0, 0.0],
    "Crystal loss (1/mm)": [0.0, 0.0, 0.0],
    "Energy/power (J or W)": (signal_E, pump_E, 0),
    "Pulse duration (fwhm ns)": (signal_ns, pump_ns, 0),
    "Beam diam. (fwhm mm)": (signal_mm, pump_mm, upconverted_mm),
    "Supergaussian coeff.": [1.0, 1.0, 1.0],
    "n2 red1 (sq cm/W)": [0.0, 0.0, 0.0],
    "n2 red2 (sq cm/W)": [0.0, 0.0, 0.0],
    "n2 blue (sq cm/W)": [0.0, 0.0, 0.0],
    "beta red1 (cm/W)": [0.0, 0.0, 0.0],
    "beta red2 (cm/W)": [0.0, 0.0, 0.0],
    "beta blue (cm/W)": [0.0, 0.0, 0.0],
    "Walkoff angles (mrad)": walkoff,
    "Offset in wo dir. (mm)": walkoff_offset,
    "Rad. curv. (mm/air)": (signal_curv_mm, pump_curv_mm, upconverted_curv_mm),
    "# of integ/grid points": [30.0, 32.0, 32.0],
    "Crystal/grid sizes (mm)": grid,
    "Deff (pm/V)": Deff,
    "Delta k (1/mm)": [0.0],
    "Dist. to image (mm)": [0.0],
    "# time steps": [20.0],
})


# %%
# Run simulation

# Parameters
inputs = np.linspace(1e-4, 15e-3, 20, endpoint=False)

utils.alt_tab()
output = []
for i in inputs:
    print(i)
    result = mix.configure_run_read({"Energy/power (J or W)": (None, i, None)})
    output.append(result["Output pulse energy (mJ)"][2])  # type: ignore
utils.alt_tab()
print(output)


# %%%
# Plot most recent values

fig = plt. figure()
ax = plt.gca()
ax.plot(inputs, output, ls=":", marker="+")

ax.set_xlabel("pump energy in mJ")
ax.set_ylabel("upconversion energy in mJ")

plt.show()



# MIT License

# Copyright (c) 2024 Benedikt Burger

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import time
from typing import Any, Optional, Protocol

from .utils import Point, gui, scale, get_content_complete, set_value
from .functions import Functions, open_function


class BaseFunction(Protocol):
    """Base class for a function window."""

    _function: Functions
    # Positions
    _run_pos: Point  # of the run field
    _result_pos: Point  # of the results field
    _close_pos: Point
    _configuration_pos: dict[str, list[Point]]  # of the configuration fields

    def open(self) -> None:
        """Open the function."""
        open_function(self._function)

    def run(self) -> None:
        """Click 'Run'."""
        gui.click(*scale(*self._run_pos))

    def close(self) -> None:
        """Click 'x'."""
        self.open()
        gui.click(*scale(*self._close_pos))

    def configure(
        self, data: Optional[dict[str, Any | list[Any] | tuple[Any, ...]]] = None
    ) -> None:
        """Configure the values and leave the config window open.

        If any value is "None", that field will not be changed. This is useful, if you want to
        change a single value in a row.
        For example `data={'Wavelengths (nm)': [1064.5, None, None]}` will set the first wavelength
        to 1064.5 nm while leaving the other wavelengths untouched.
        """
        self.open()
        if data is None:
            return
        for key, value in data.items():
            positions = self._configuration_pos[key]
            if isinstance(value, (list, tuple)):
                for i, val in enumerate(value):
                    if val is not None:
                        set_value(positions[i], val)
            else:
                if value is not None:
                    set_value(positions[0], value)

    def get_configuration(self) -> dict[str, list[float | str]]:
        """Read the current configuration."""
        self.open()
        data = {}
        for key, positions in self._configuration_pos.items():
            d = []
            for pos in positions:
                val = get_content_complete(pos)
                try:
                    d.append(float(val))
                except ValueError:
                    d.append(val)
            data[key] = d
        return data

    def read_results(self) -> list[str]:
        return get_content_complete(self._result_pos).split("\r\n")

    def interpret_results(self, rows: list[str]) -> dict[str, Any]:
        """Interpret the results and return them as a dictionary."""
        data = {}
        for row in rows:
            # skip empty lines
            if not row:
                continue
            if row == 'ERROR: No phase match found.':
                raise Exception('No phase match found')
            
            text, values = row.split("=")
            content = []
            for element in values.split():
                try:
                    val = float(element)
                except ValueError:
                    val = element
                content.append(val)
            data[text.strip()] = content
        return data

    def run_and_read(
        self,
        waiting_time: float = 1,
        max_tries: int = 10,
        interval: float = 0.5,
        waiting_line_count: int = 3,
    ) -> dict[str, Any]:
        """Run an analysis and return the result."""
        self.run()
        time.sleep(waiting_time)
        for _ in range(max_tries):
            rows = self.read_results()
            if len(rows) > waiting_line_count:
                break
            time.sleep(interval)

        # interpret results and save as dictionary:
        return self.interpret_results(rows)

    def configure_run_read(
        self, data: Optional[dict[str, Any]] = None, **kwargs
    ) -> dict[str, Any]:
        """Configure and run an analysis and return the result."""
        self.configure(data)
        return self.run_and_read(**kwargs)


def generate_position_dict(
    first_position: Point,
    configuration_names: list[str],
    columns: int = 3,
    column_distance: int = 60,
) -> dict[str, list[Point]]:
    """Generate a position dictionary for functions.

    This utility makes it easier to generate a position matrix with three fields per entry.
    Adjust it afterwards according to the number of entries.

    :param first_position: Position (tuple) of the top left configuration field.
    :param configuration_names: List of all the names (in order) of the configuration.
    """
    # Notes regarding positions for mix: horizontal distance is 60 px, vertical distance 16 pixels
    positions = {}
    i = 0
    for name in configuration_names:
        positions[name] = [
            (first_position[0] + j * column_distance, first_position[1] + 16 * i)
            for j in range(columns)
        ]
        i += 1
    return positions



# MIT License

# Copyright (c) 2024 Benedikt Burger

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging

from . import utils

from .main_window import open_function, Functions  # noqa: F401
from .ref_index import RefractiveIndex  # noqa: F401
from .two_d_mix_lp import TwoDMixLP  # noqa: F401
from .two_d_mix_sp import TwoDMixSP  # noqa: F401
from .focus import focus  # noqa: F401

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def main() -> None:
    """Initialize the SNLO helper."""
    if len(log.handlers) < 2:
        log.addHandler(logging.StreamHandler())
    log.setLevel(logging.INFO)
    utils.set_screenfactors()
    log.info(f"Setting screenfactors to {utils.factors}.")


if __name__ == "__main__":  # pragma: nocover
    main()



# MIT License

# Copyright (c) 2024 Benedikt Burger

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Example on how to use the SNLO data files.
"""


import numpy as np
import time
from typing import cast

import matplotlib.pyplot as plt
import matplotlib.ticker as tic
import pyautogui as gui
from scipy import constants as cs
from scipy.optimize import curve_fit

from snlohelper.utils import scale
from snlohelper.import_snlo_file import import_snlo_file


def gauss(x, x0, a, sigma):
    """Gaussian function"""
    return a * np.exp(-((x - x0) ** 2) / (2 * sigma**2))


def plotdata(data, xScale=1, yScale=1):
    """Returns the x and y components of a two dimensional list as individual lists, e.g., for
    plotting. Optional linear scale factors as additional commands"""
    xValues = []
    yValues = []
    for i in range(len(data)):
        xValues.append(xScale * data[i][0])
        yValues.append(yScale * data[i][1])
    return xValues, yValues


def gaussian_fit(data, print_output=False):
    """performs Gaussian fit (position, amplitude, standard deviation), returns popt, sdev, serr,
    pcov"""
    # arguments of gauss-function: gauss(x, x0, a, sigma)
    x, y = plotdata(data)

    # find peak position
    peak_pos = x[list(y).index(max(y))]

    try:
        popt, pcov = curve_fit(gauss, x, y, p0=[peak_pos, 1, 1])
    except RuntimeError:
        print("fit failed!")
        popt, pcov = [0, 0, 0], np.zeros((3, 3)).tolist()

    # covariance matrix: diagonal elements are the variances, hence, the square root is the
    # standard deviation
    sdev = np.sqrt(abs(pcov.diagonal()).tolist())

    # standard errors are sd/sqrt(n) with the sample size n
    serr = sdev / (np.sqrt(len(data)))

    return popt, sdev, serr, pcov


def plot_options(
    fig,
    ax,
    xlabel=None,
    ylabel=None,
    font_size=11,
    ticks=["auto", "auto"],
    image_width=10,
    aspect_ratio=cs.golden,
    dpi=200,
):
    """
    sets plot options just like in Mathematica
    :param fig: figure object from pyplot (necessary)
    :param ax: axes object from pyplot (necessary)
    :param xlabel: x-axis label (string or None)
    :param ylabel: y-axis label (string or None)
    :param font_size: font size (number, standard=11)
    :param ticks: custom or auto ticks (list of x and y: 'auto' or major tick increment)
    :param image_width: width of the image in cm (number, standard=10)
    :param aspect_ratio: aspect ratio of the image (number, standard=golden ratio)

    Parameters
    ----------
    dpi : dpi for displaying in jupyterlab (standard = 200)
    """
    # set size and aspect ratio
    fig.set_size_inches(image_width / 2.54, image_width / 2.54 / aspect_ratio)

    fig = plt.gcf()
    plt.rcParams["font.family"] = "Arial"
    fig.set_dpi(dpi)  # only for display in jupyterlab!
    fig.patch.set_facecolor("white")  # type: ignore

    # plt.style.use('classic')
    ax.tick_params(
        axis="both", which="major", labelsize=font_size, right=True, top=True, direction="in"
    )
    ax.tick_params(
        axis="both", which="minor", labelsize=font_size - 2, right=True, top=True, direction="in"
    )

    # custom major ticks
    if ticks[0] == "auto":
        ax.xaxis.set_major_locator(tic.AutoLocator())
    else:
        ax.xaxis.set_major_locator(tic.MultipleLocator(ticks[0]))
    if ticks[1] == "auto":
        ax.yaxis.set_major_locator(tic.AutoLocator())
    else:
        ax.yaxis.set_major_locator(tic.MultipleLocator(ticks[1]))

    # activate automatic minor ticks
    ax.xaxis.set_minor_locator(tic.AutoMinorLocator())
    ax.yaxis.set_minor_locator(tic.AutoMinorLocator())

    # changes fontsize of the labels
    ax.set_xlabel(xlabel, color="black", fontsize=font_size)
    ax.set_ylabel(ylabel, color="black", fontsize=font_size)

    # reduce tick lengths
    ax.tick_params(axis="both", which="minor", length=2)
    ax.tick_params(axis="both", which="major", length=3)


def get_spectr(
    call_function: bool = True, display_results: bool = False
) -> dict[str, float]:
    """imports spectra and power vs. time data for all three fields of the mixing process.
    call_function: for 2D-mix-LP you have to click on 'Spectra' to calculate the spectra.
    display_results: gives plots for spectra and powers
    returns dictionary with bandwidths and durations (given as FWHMs)"""
    if call_function:
        gui.click(*scale(400, 260))
        time.sleep(0.5)

    # detuning [MHz], Red1, Red2, Blue
    spectr = import_snlo_file("OPA2D_SP.dat")
    # time, power, phase, Mx^2, My^2, Rad Curv c, Rad Curv y, X-tilt, w_x^2, w_y^2
    id_beam = import_snlo_file("ID_BEAM.dat")
    sig_beam = import_snlo_file("SIG_BEAM.dat")
    pmp_beam = import_snlo_file("PMP_BEAM.dat")

    # fit idler duration
    x = 1e9 * id_beam.T[0]  # ns
    y = id_beam.T[1]
    idl = gaussian_fit(np.array([x, y]).T, False)[0]
    fwhm_ns_idl = abs(2 * np.sqrt(2 * np.log(2)) * idl[2])
    x_lists = [x.copy()]
    y_lists = [y.copy()]

    # fit signal duration
    x = 1e9 * sig_beam.T[0]  # ns
    y = sig_beam.T[1]
    sig = gaussian_fit(np.array([x, y]).T, False)[0]
    fwhm_ns_sig = abs(2 * np.sqrt(2 * np.log(2)) * sig[2])
    x_lists.append(x)
    y_lists.append(y)

    # fit pump duration
    x = 1e9 * pmp_beam.T[0]  # ns
    y = pmp_beam.T[1]
    pmp = gaussian_fit(np.array([x, y]).T, False)[0]
    fwhm_ns_pmp = abs(2 * np.sqrt(2 * np.log(2)) * pmp[2])
    x_lists.append(x)
    y_lists.append(y)
    fits = idl, sig, pmp
    if display_results:
        # plot pulse durations
        fig, ax = plt.subplots(1, 3)
        fwhms = fwhm_ns_idl, fwhm_ns_sig, fwhm_ns_pmp
        titles = "idler (Red1)", "signal (Red2)", "pump (Blue)"
        colors = "red", "green", "blue"
        for i in range(len(ax)):
            plot_options(
                fig, ax[i], xlabel="time (ns)", image_width=15, aspect_ratio=3
            )
            ax[i].plot(x_lists[i], y_lists[i], ".", color=colors[i])
            ax[i].set_title(titles[i], color=colors[i])

            x_fit = np.arange(min(x_lists[i]), max(x_lists[i]), 0.1)

            ax[i].plot(x_fit, gauss(x_fit, *fits[i]), "-", color=colors[i])
            ax[i].text(0, 0, f"FWHM\n{fwhms[i]:.1f}ns", ha="center", color=colors[i])
            # ax[i].set_xlim([min(x_lists[i]), max(x_lists[i])])
        plot_options(
            fig,
            ax[0],
            image_width=15,
            aspect_ratio=3,
            xlabel="time (ns)",
            ylabel="normalized power",
        )
        plt.tight_layout()
        plt.show()

    # fit spectra
    x = spectr.T[0]  # in MHz
    y_idl = spectr.T[1]
    y_sig = spectr.T[2]
    y_pmp = spectr.T[3]

    idl = cast(list[float], gaussian_fit(np.array([x, y_idl]).T, False)[0])
    sig = cast(list[float], gaussian_fit(np.array([x, y_sig]).T, False)[0])
    pmp = cast(list[float], gaussian_fit(np.array([x, y_pmp]).T, False)[0])

    fwhm_spec_idl = abs(2 * np.sqrt(2 * np.log(2)) * idl[2])
    fwhm_spec_sig = abs(2 * np.sqrt(2 * np.log(2)) * sig[2])
    fwhm_spec_pmp = abs(2 * np.sqrt(2 * np.log(2)) * pmp[2])

    y_lists = y_idl, y_sig, y_pmp

    fits = idl, sig, pmp
    if display_results:
        # plot spectra durations
        fig, ax = plt.subplots(1, 3)
        fwhms = fwhm_spec_idl, fwhm_spec_sig, fwhm_spec_pmp
        titles = "idler (Red1)", "signal (Red2)", "pump (Blue)"
        colors = "red", "green", "blue"
        for i in range(len(ax)):
            plot_options(
                fig, ax[i], xlabel="detuning (MHz)", image_width=15, aspect_ratio=3
            )
            ax[i].plot(x, y_lists[i], ".", color=colors[i])
            ax[i].set_title(titles[i], color=colors[i])

            x_fit = np.arange(min(x), max(x), 0.1)

            ax[i].plot(x_fit, gauss(x_fit, *fits[i]), "-", color=colors[i])
            ax[i].text(0, 0, f"FWHM\n{fwhms[i]:.1f}MHz", ha="center", color=colors[i])
            ax[i].set_xlim([-2 * fwhms[i], 2 * fwhms[i]])
        plot_options(
            fig,
            ax[0],
            image_width=15,
            aspect_ratio=3,
            xlabel="time (ns)",
            ylabel="normalized power",
        )
        plt.tight_layout()
        plt.show()

    return {
        "fwhm_idl_MHz": fwhm_spec_idl,
        "fwhm_sig_MHz": fwhm_spec_sig,
        "fwhm_pmp_MHz": fwhm_spec_pmp,
        "duration_idl_ns": fwhm_ns_idl,
        "duration_sig_ns": fwhm_ns_sig,
        "duration_pmp_ns": fwhm_ns_pmp,
    }



# MIT License

# Copyright (c) 2024 Benedikt Burger

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, Optional, Protocol

from .utils import Point, gui, scale
from .base_function import BaseFunction


class MixMethods(BaseFunction, Protocol):
    """Parent class for mix methods.

    Subclass it for specific methods. You should define the positions and the result interpretation.
    """

    _accept_pos: Point
    _change_inputs_pos: Point

    def accept(self) -> None:
        """Click 'Accept'."""
        gui.click(*scale(*self._accept_pos))

    def change_inputs(self) -> None:
        """Click 'Change Inputs'."""
        gui.click(*scale(*self._change_inputs_pos))

    def configure_run_read(
        self, data: Optional[dict[str, Any]] = None, **kwargs
    ) -> dict[str, float | list[float]]:
        """Configure and run an analysis and return the result."""
        self.configure(data)
        self.accept()
        return self.run_and_read(**kwargs)
